/**
  *  \file sstring.cc
  */

#include <cstring>
#include "cpluslib/sstring.h"
#include "cpluslib/assert.h"
#include "cpluslib/cstring.h"
#include "cpluslib/substr.h"

/** Allocation Policy. */
enum {
    /** Above ALLOC_THRESHOLD, we add ALLOC_INC. Below, we double. */
    SSTRING_ALLOC_THRESHOLD = 8191,
    SSTRING_ALLOC_INC       = 128,
    /** We allocate no less than this. */
    SSTRING_ALLOC_MIN       = 15,
};

/** \class simple_string
    \brief Simple, Small String

    A string class which, unlike std::string, generates very tight
    code for most operations by avoiding temporaries with destructors. 
    These generate a lot of overhead for the unlikely case of an
    exception (test suite with simple_string: 30k, with std::string: 45k).

    This class is otherwise (intended to be) exactly like std::string,
    with the following exceptions:
    - it's not a template. Rationale: allows putting implementation code
      in a .cc file.
    - allocators are not implemented. Rationale: no templates.
    - certain operations which require simple_string objects do not
      work with simple_string expressions because of the proxy objects
      (e.g. a `class foo' with a ctor `foo::foo(simple_string)',
      and a function `foothing(foo)', the function can not be
      called as `foothing(a + b)'; with std::string, this would work.
      Workaround: add a cast: `foothing(simple_string(a + b))'
    - a simple_string must not be modified while an expression referring
      to it is being evaluated. I.e. `(x+="a") + (x+="b") + (x+="c")'
      does not work. This problem can be solved (by having the proxy
      objects point to the actual string objects and not their contents)
      but I think it's not worth it: `(i+=1) + (i+=2) + (i+=3)' with
      i being int has undefined behaviour, and nobody complains there.
    - this implementation does not fulfill the STL namespace rules ;-)

    Implementation note: we always allocate one byte more than needed,
    to support an efficient/simple version of c_str(). When the string
    is empty, txt may be null. */

/*
  To do:
  + operator>>
  + optimisations:
    - compare(proxy)
    - append(proxy), operator+(string, proxy)
  + reverse iterator
  + do we need more test cases?
  + simple_stringbuf / simple_stringstream
  + length overflow not yet handled (i.e. taking a 1G string and doing
    `s + s + s')
  + inline some more functions?
 */

/*
  Guidelines:
  + Use C functions, not <algorithm>. Rationale: no exceptions, speed
    (gcc automatically knows that C functions never throw)
 */

const char simple_string::null = 0;

template<class T>
static inline T
xstr_min(T a, T b)
{
    return a < b ? a : b;
}

/*********************** simple_string constructors **********************/

/* copy ctor */
simple_string::simple_string(const simple_string& rhs)
{
    if (rhs.len) {
        cap = len = rhs.len;
        txt = new charT[rhs.len+1];
        std::memcpy(txt, rhs.txt, len);
    } else {
        cap = len = 0;
        txt = 0;
    }
}

/* init with C string */
simple_string::simple_string(const charT* src)
{
    if (*src) {
        cap = len = std::strlen(src);
        txt = new charT[len+1];
        std::memcpy(txt, src, len);
    } else {
        cap = len = 0;
        txt = 0;
    }
}

/* init with str.substr(pos, n) */
simple_string::simple_string(const simple_string& str, size_type pos,
                             size_type n)
{
    ASSERT(pos <= str.length());

    cap = len = xstr_min(str.length() - pos, n);
    if (len) {
        txt = new charT[len+1];
        std::memcpy(txt, str.txt + pos, len);
    } else {
        txt = 0;
    }
}

simple_string::simple_string(const substring& str)
{
    if (str.length()) {
        cap = len = str.length();
        txt = new charT[len+1];
        std::memcpy(txt, str.data_or_null(), len);
    } else {
        cap = len = 0;
        txt = 0;
    }
}

/* init with [src, src+n) */
simple_string::simple_string(const charT* src, size_type n)
    : txt(new charT[n+1]), len(n), cap(n)
{
    std::memcpy(txt, src, n);
}

/* init with n copies of c */
simple_string::simple_string(size_type n, charT c)
    : txt(new charT[n+1]), len(n), cap(n)
{
    std::memset(txt, c, n);
}

/*********************** init-from-proxy functions ***********************/

void
simple_string::init_from_proxy(const proxy& rhs)
{
    sizeT newlen = rhs.length();

    if (newlen) {
        txt = new charT[newlen+1];
        cap = len = newlen;
        for (const proxy* p = &rhs; p; p = p->next)
            if (p->len)
                std::memcpy(&txt[newlen -= p->len], p->txt, p->len);
    } else {
        txt = 0;
        cap = len = 0;
    }
}

void
simple_string::assign_from_proxy(const proxy& rhs)
{
    sizeT newlen = rhs.length();
    
    if (newlen) {
        charT* ntxt = new charT[newlen+1];
        sizeT nlen = newlen;
        for (const proxy* p = &rhs; p; p = p->next)
            if (p->len)
                std::memcpy(&ntxt[newlen -= p->len], p->txt, p->len);
        delete[] txt;
        txt = ntxt;
        cap = len = nlen;
    } else {
        delete[] txt;
        txt = 0;
        cap = len = 0;
    }
}

/***************************** erase methods *****************************/

simple_string&
simple_string::erase(size_type pos, size_type n)
{
    replace(pos, n, (charT*) 0, 0);
    return *this;
}

simple_string::iterator
simple_string::erase(iterator position)
{
    size_type pos = position - begin();
    replace(pos, 1, (charT*) 0, 0);
    return begin() + pos;
}

simple_string::iterator
simple_string::erase(iterator first, iterator last)
{
    size_type pos = first - begin();
    replace(pos, last - first, (charT*) 0, 0);
    return begin() + pos;
}

/**************************** replace methods ****************************/

simple_string&
simple_string::replace(size_type pos1, size_type n1,
                       const simple_string& str)
{
    return replace(pos1, n1, str.data_or_null(), str.length());
}

simple_string&
simple_string::replace(size_type pos1, size_type n1,
                       const simple_string& str,
                       size_type pos2, size_type n2)
{
//     if (n2 > str.length() - pos2)
//         n2 = str.length() - pos2;
    n2 = xstr_min(n2, str.length() - pos2);

    return replace(pos1, n1, str.data() + pos2, n2);
}

simple_string&
simple_string::replace(size_type pos, size_type n1,
                       const charT* s, size_type n2)
{
    ASSERT(pos <= length());

    if (n1 > length() - pos)
        n1 = length() - pos;

    if (len + n2 - n1 <= cap && (s < txt || s > txt + len)) {
        if (length() - pos - n1 && n1 != n2)
            std::memmove(txt + pos + n2, txt + pos + n1, length() - pos - n1);
        std::memcpy(txt + pos, s, n2);
        len = len + n2 - n1;
    } else {
        proxy2 px(0, data_or_null(), pos, s, n2);
        assign_from_proxy(proxy(&px, data_or_null() + pos + n1, length() - pos - n1));
    }
    return *this;
}

simple_string&
simple_string::replace(size_type pos, size_type n1, const charT* s)
{
    return replace(pos, n1, s, std::strlen(s));
}

simple_string&
simple_string::replace(size_type pos, size_type n1, size_type n2, charT c)
{
    return replace(pos, n1, simple_string(n2, c));
}

simple_string&
simple_string::replace(iterator i1, iterator i2, const simple_string& str)
{
    return replace(i1 - begin(), i2 - i1, str);
}

simple_string&
simple_string::replace(iterator i1, iterator i2, const charT* s, size_type n)
{
    return replace(i1 - begin(), i2 - i1, s, n);
}

simple_string&
simple_string::replace(iterator i1, iterator i2, const charT* s)
{
    return replace(i1 - begin(), i2 - i1, s);
}

simple_string&
simple_string::replace(iterator i1, iterator i2, size_type n, charT c)
{
    return replace(i1 - begin(), i2 - i1, n, c);
}

/***************************** insert methods ****************************/

simple_string&
simple_string::insert(size_type pos1, const simple_string& str)
{
    return replace(pos1, 0, str);
}

simple_string&
simple_string::insert(size_type pos1, const simple_string& str,
                      size_type pos2, size_type n)
{
    return replace(pos1, 0, str, pos2, n);
}

simple_string&
simple_string::insert(size_type pos, const charT* s, size_type n)
{
    return replace(pos, 0, s, n);
}

simple_string&
simple_string::insert(size_type pos, const charT* s)
{
    return replace(pos, 0, s);
}

simple_string&
simple_string::insert(size_type pos, size_type n, charT c)
{
    return replace(pos, 0, n, c);
}

simple_string::iterator
simple_string::insert(iterator p, charT c)
{
    size_type pos = p - begin();
    insert(pos, 1, c);
    return begin() + pos;
}

void
simple_string::insert(iterator p, size_type n, charT c)
{
    insert(p - begin(), n, c);
}

/**************************** compare methods ****************************/

int
simple_string::compare(const simple_string& str) const
{
    // return compare(0, npos, str.data_or_null(), str.length());
    return cstring_util::compare(data_or_null(), length(),
                                 str.data_or_null(), str.length());
}

int
simple_string::compare(size_type pos1, size_type n1,
                       const simple_string& str) const
{
    // return compare(pos1, n1, str.data_or_null(), str.length());
    return cstring_util::compare(data_or_null() + pos1,
                                 xstr_min(length()-pos1, n1),
                                 str.data_or_null(),
                                 str.length());
}

int
simple_string::compare(size_type pos1, size_type n1,
                       const simple_string& str,
                       size_type pos2, size_type n2) const
{
//     if (n2 > str.length() - pos2)
//         n2 = str.length() - pos2;
//     return compare(pos1, n1, str.data_or_null() + pos2, n2);
    return cstring_util::compare(data_or_null() + pos1,
                                 xstr_min(length()-pos1, n1),
                                 str.data_or_null() + pos2,
                                 xstr_min(str.length()-pos2, n2));
}

int
simple_string::compare(const charT* s) const
{
//     return compare(0, npos, s, std::strlen(s));
    return cstring_util::compare(data_or_null(), length(),
                                 s, std::strlen(s));
}

int
simple_string::compare(size_type pos1, size_type n1,
                       const charT* s, size_type n2) const
{
//     if (n1 > length() - pos1)
//         n1 = length() - pos1;
    
//     int cmplen = xstr_min(n1, n2);
//     int mres = cmplen ? std::memcmp(data() + pos1, s, cmplen) : 0;
//     if (!mres)
//         return n1 > n2 ? 1 : n1 < n2 ? -1 : 0;
//     else
//         return mres;
    return cstring_util::compare(data_or_null() + pos1, xstr_min(length()-pos1, n1), s, n2);
}

/****************************** find methods *****************************/

simple_string::size_type
simple_string::find (const simple_string& str, size_type pos) const
{
//     return find(str.data_or_null(), pos, str.length());
    return cstring_util::find(data_or_null(), length(), pos,
                              str.data_or_null(), str.length());
}

simple_string::size_type
simple_string::find (const charT* s, size_type pos, size_type n) const
{
    return cstring_util::find(data_or_null(), length(), pos, s, n);
//     if (pos >= length() || n > length() - pos)
//         return npos;
//     if (!n)
//         return pos;

//     size_type maxpos = length() - n;
//     while (pos <= maxpos && (pos = find(*s, pos)) != npos && pos <= maxpos) {
//         if (compare(pos, n, s, n) == 0)
//             return pos;
//         ++pos;
//     }
//     return npos;
}

simple_string::size_type
simple_string::find (const charT* s, size_type pos) const
{
//     return find(s, pos, std::strlen(s));
    return cstring_util::find(data_or_null(), length(), pos, s, std::strlen(s));
}

simple_string::size_type
simple_string::find (charT c, size_type pos) const
{
//     if (pos >= length())
//         return npos;

//     charT* found = (charT*) std::memchr(txt + pos, c, len - pos);
//     if (found)
//         return found - txt;
//     else
//         return npos;
    return cstring_util::find(data_or_null(), length(), pos, c);
}

/***************************** rfind methods *****************************/

simple_string::size_type
simple_string::rfind(const simple_string& str, size_type pos) const
{
//     return rfind(str.data(), pos, str.size());
    return cstring_util::rfind(data_or_null(), length(), pos,
                               str.data_or_null(), str.length());
}

simple_string::size_type
simple_string::rfind(const charT* s, size_type pos, size_type n) const
{
//     /* cd2 doesn't say what to do if n==0 */
//     if (n > length() || !n)
//         return npos;
    
//     if (pos >= length() - n)
//         pos = length() - n;

//     while ((pos = rfind(*s, pos)) != npos) {
//         if (compare(pos, n, s, n) == 0)
//             return pos;
//         if (!pos)
//             return npos;
//         --pos;
//     }
//     return npos;
    return cstring_util::rfind(data_or_null(), length(), pos, s, n);
}

simple_string::size_type
simple_string::rfind(const charT* s, size_type pos) const
{
//     return rfind(s, pos, std::strlen(s));
    return cstring_util::rfind(data_or_null(), length(), pos, s, std::strlen(s));
}

simple_string::size_type
simple_string::rfind(charT c, size_type pos) const
{
//     if (pos > size())
//         pos = size();
//     if (pos < size() && txt[pos] == c)
//         return pos;
//     while (pos != 0) {
//         if (txt[--pos] == c)
//             return pos;
//     }
//     return npos;
    return cstring_util::rfind(data_or_null(), length(), pos, c);
}

/************************* find_first_of methods *************************/

simple_string::size_type
simple_string::find_first_of(const simple_string& str, size_type pos) const
{
//     return find_first_of(str.data_or_null(), pos, str.length());
    return cstring_util::find_first_of(data_or_null(), length(), pos,
                                       str.data_or_null(), str.length());
}

simple_string::size_type
simple_string::find_first_of(const charT* s, size_type pos, size_type n) const
{
//     if (!n)
//         return npos;

//     while (pos < length()) {
//         if (std::memchr(s, txt[pos], n))
//             return pos;
//         ++pos;
//     }
//     return npos;
    return cstring_util::find_first_of(data_or_null(), length(), pos,
                                       s, n);
}

simple_string::size_type
simple_string::find_first_of(const charT* s, size_type pos) const
{
//     return find_first_of(s, pos, std::strlen(s));
    return cstring_util::find_first_of(data_or_null(), length(), pos,
                                       s, std::strlen(s));
}

simple_string::size_type
simple_string::find_first_of(charT c, size_type pos) const
{
    return find(c, pos);
}

/************************** find_last_of methods *************************/

simple_string::size_type
simple_string::find_last_of (const simple_string& str, size_type pos) const
{
//     return find_last_of(str.data_or_null(), pos, str.size());
    return cstring_util::find_last_of(data_or_null(), length(), pos,
                                      str.data_or_null(), str.length());
}

simple_string::size_type
simple_string::find_last_of (const charT* s, size_type pos, size_type n) const
{
//     if (!n)
//         return npos;
    
//     if (pos > length())
//         pos = length();

//     if (pos < length() && std::memchr(s, txt[pos], n))
//         return pos;
//     while (pos) {
//         if (std::memchr(s, txt[--pos], n))
//             return pos;
//     }
//     return npos;
    return cstring_util::find_last_of(data_or_null(), length(), pos,
                                      s, n);
}

simple_string::size_type
simple_string::find_last_of (const charT* s, size_type pos) const
{
//     return find_last_of(s, pos, std::strlen(s));
    return cstring_util::find_last_of(data_or_null(), length(), pos,
                                      s, std::strlen(s));
}

simple_string::size_type
simple_string::find_last_of (charT c, size_type pos) const
{
    return rfind(c, pos);
}

/*********************** find_first_not_of methods ***********************/

simple_string::size_type
simple_string::find_first_not_of(const simple_string& str, size_type pos) const
{
    return cstring_util::find_first_not_of(data_or_null(), length(), pos,
                                           str.data_or_null(), str.length());
//     return find_first_not_of(str.data_or_null(), pos, str.length());
}

simple_string::size_type
simple_string::find_first_not_of(const charT* s, size_type pos, size_type n) const
{
//     if (!n)
//         return xstr_min(pos, length());
    
//     while (pos < length()) {
//         if (! std::memchr(s, txt[pos], n))
//             return pos;
//         ++pos;
//     }
//     return npos;
    return cstring_util::find_first_not_of(data_or_null(), length(), pos,
                                           s, n);
}

simple_string::size_type
simple_string::find_first_not_of(const charT* s, size_type pos) const
{
//     return find_first_not_of(s, pos, std::strlen(s));
    return cstring_util::find_first_not_of(data_or_null(), length(), pos,
                                           s, std::strlen(s));
}

simple_string::size_type
simple_string::find_first_not_of(charT c, size_type pos) const
{
    return cstring_util::find_first_not_of(data_or_null(), length(), pos,
                                           &c, 1);
}

/************************ find_last_not_of methods ***********************/

simple_string::size_type
simple_string::find_last_not_of (const simple_string& str, size_type pos) const
{
//     return find_last_not_of(str.data_or_null(), pos, str.size());
    return cstring_util::find_last_not_of(data_or_null(), length(), pos,
                                          str.data_or_null(), str.length());
}

simple_string::size_type
simple_string::find_last_not_of (const charT* s, size_type pos, size_type n) const
{
//     if (!n)
//         return pos;             // FIXME?
    
//     if (pos > length())
//         pos = length();

//     if (pos < length() && ! std::memchr(s, txt[pos], n))
//         return pos;
//     while (pos) {
//         if (! std::memchr(s, txt[--pos], n))
//             return pos;
//     }
//     return npos;
    return cstring_util::find_last_not_of(data_or_null(), length(), pos,
                                          s, n);
}

simple_string::size_type
simple_string::find_last_not_of (const charT* s, size_type pos) const
{
//     return find_last_not_of(s, pos, std::strlen(s));
    return cstring_util::find_last_not_of(data_or_null(), length(), pos,
                                          s, std::strlen(s));
}

simple_string::size_type
simple_string::find_last_not_of (charT c, size_type pos) const
{
//     return find_last_not_of(&c, pos, 1);
    return cstring_util::find_last_not_of(data_or_null(), length(), pos,
                                          &c, 1);
}

/****************************** misc methods *****************************/

void
simple_string::resize(simple_string::size_type n, charT c)
{
    if (n < len)
        assign_from_proxy(proxy(0, txt, n));
    else if (n > len)
        append(n - len, c);
}

void
simple_string::reserve(size_type n)
{
    delete ensure_capacity(n);
}

void
simple_string::clear()
{
    if (len || cap) {
        delete[] txt;
        txt = 0;
        cap = len = 0;
    }
}

simple_string::size_type
simple_string::copy(charT* s, size_type n, size_type pos) const
{
    ASSERT(pos <= length());
    size_type rlen = xstr_min(n, length() - pos);
    std::memcpy(s, data_or_null() + pos, rlen);
    return rlen;
}

simple_string::reference
simple_string::at(size_type pos)
{
//     if (pos >= len)
//         throw std::out_of_range("invalid string position");
    ASSERT(pos < len);
    return txt[pos];
}

simple_string::const_reference
simple_string::at(size_type pos) const
{
//     if (pos >= len)
//         throw std::out_of_range("invalid string position");
    ASSERT(pos < len);
    return txt[pos];
}

simple_string::proxy
simple_string::substr(size_type pos, size_type n) const
{
    ASSERT(pos <= length());
    if (n > length() - pos)
        n = length() - pos;
    if (n)
        return proxy(0, txt + pos, n);
    else
        return proxy(0, 0, 0);
}

void
simple_string::swap(simple_string& rhs)
{
    charT* txt_save = txt;
    sizeT len_save = len;
    sizeT cap_save = cap;
    txt = rhs.txt;
    len = rhs.len;
    cap = rhs.cap;
    rhs.txt = txt_save;
    rhs.len = len_save;
    rhs.cap = cap_save;
}

simple_string::charT*
simple_string::ensure_capacity(sizeT newcap)
{
    if (newcap <= cap)
        return 0;

    sizeT nc = cap;
    if (nc < SSTRING_ALLOC_MIN)
        nc = SSTRING_ALLOC_MIN;
    while (nc < newcap) {
        if (nc < SSTRING_ALLOC_THRESHOLD)
            nc *= 2;
        else
            nc += SSTRING_ALLOC_INC;
    }
    charT* nb = new char[nc+1];
    charT* rv = txt;
    if (len)
        std::memcpy(nb, txt, len);
    txt = nb;
    cap = nc;
    return rv;
}

/***************************** append methods ****************************/

simple_string&
simple_string::append(const simple_string& str)
{
    // assign_from_proxy(proxy2(0, data(), length(), str.data(), str.length()));
    append(str.data_or_null(), str.size());
    return *this;
}

simple_string&
simple_string::append(const simple_string& str, size_type pos, size_type n)
{
    ASSERT(pos <= str.length());

    size_type rest = xstr_min(n, str.length() - pos);
    // assign_from_proxy(proxy2(0, data(), length(), str.data() + pos, rest));
    append(str.data_or_null() + pos, rest);
    return *this;
}

simple_string&
simple_string::append(const charT* s, size_type n)
{
//     assign_from_proxy(proxy2(0, data(), length(), s, n));
    if (n) {
        charT* dm = ensure_capacity(length() + n);
        std::memcpy(txt + len, s, n);
        len += n;
        delete[] dm;
    }
    return *this;
}

simple_string&
simple_string::append(const charT* s)
{
    // assign_from_proxy(proxy2(0, data(), length(), s, std::strlen(s)));
    append(s, std::strlen(s));
    return *this;
}

simple_string&
simple_string::append(size_type n, charT c)
{
    charT* dm = ensure_capacity(length() + n);
    std::memset(txt + len, c, n);
    len += n;
    delete[] dm;
    return *this;
}

/***************************** assign methods ****************************/

simple_string&
simple_string::assign(const simple_string& str)
{
    assign_from_proxy(proxy(0, str.data_or_null(), str.length()));
    return *this;
}

simple_string&
simple_string::assign(const simple_string& str, size_type pos,
                      size_type n)
{
    ASSERT(pos <= str.length());

    size_type rest = xstr_min(n, str.length() - pos);
    assign_from_proxy(proxy(0, str.data_or_null() + pos, rest));
    return *this;
}

simple_string&
simple_string::assign(const charT* s, size_type n)
{
    assign_from_proxy(proxy(0, s, n));
    return *this;
}

simple_string&
simple_string::assign(const charT* s)
{
    assign_from_proxy(proxy(0, s, std::strlen(s)));
    return *this;
}

simple_string&
simple_string::assign(size_type n, charT c)
{
    charT* v = ensure_capacity(n);
    std::memset(txt, c, n);
    len = n;
    delete[] v;
    return *this;
}

/****************************** operator += ******************************/

simple_string&
simple_string::operator+=(const proxy& rhs)
{
    return append(simple_string(rhs));
}

/******************************* operator + ******************************/

/* string + string */
simple_string::proxy2
operator+(const simple_string& lhs, const simple_string& rhs)
{
    return simple_string::proxy2(0, lhs.data_or_null(), lhs.length(),
                                 rhs.data_or_null(), rhs.length());
}

simple_string::proxy
operator+(const simple_string::proxy& lhs, const simple_string& rhs)
{
    return simple_string::proxy(&lhs, rhs.data_or_null(), rhs.length());
}

/* string + char array */
simple_string::proxy2
operator+(const simple_string& lhs, const simple_string::charT* rhs)
{
    return simple_string::proxy2(0, lhs.data_or_null(), lhs.length(),
                                 rhs, std::strlen(rhs));
}

simple_string::proxy
operator+(const simple_string::proxy& lhs, const simple_string::charT* rhs)
{
    return simple_string::proxy(&lhs, rhs, std::strlen(rhs));
}

/* char array + string */
simple_string::proxy2
operator+(const simple_string::charT* lhs, const simple_string& rhs)
{
    return simple_string::proxy2(0, lhs, std::strlen(lhs),
                                 rhs.data_or_null(), rhs.length());
}

/* for completeness: char + string, string + char */
simple_string
operator+(simple_string::charT lhs, const simple_string& rhs)
{
    return simple_string(1, lhs) + rhs;
}

simple_string
operator+(const simple_string& lhs, simple_string::charT rhs)
{
    return lhs + simple_string(1, rhs);
}

/************************** simple_string::proxy *************************/

simple_string::size_type
simple_string::proxy::length() const
{
    size_type n = 0;
    for (const proxy* p = this; p != 0; p = p->next)
        n += p->len;
    return n;
}
